from typing import Any, Dict, Tuple

from process_bigraph import Emitter

import numpy as np
import pandas as pd
from simulariumio import (
    TrajectoryConverter,
    TrajectoryData,
    AgentData,
    MetaData,
    UnitData,
)


class SimulariumEmitter(Emitter):
    config_schema = {
        "emit": "schema",
        "output_dir": "string"
    }

    def __init__(self, config, core):
        super().__init__(config, core)
        self.configuration_data = None
        self.saved_data: Dict[float, Dict[str, Any]] = {}

    def update(self, state) -> Dict:
        if "particles" in state and "topologies" in state:
            self.saved_data[state['global_time']] = state
        return {}

    def query(self, query=None):
        output = self.config["output_dir"]
        return {'result' : self.save_simularium_file(output)}

    def get_simularium_monomers(
        self, time, monomers, actin_radius, position_offset, trajectory
    ):
        """
        Shape monomer state data into Simularium agents
        """
        time_index = len(trajectory["times"])
        trajectory["times"].append(time)
        trajectory["unique_ids"].append([])
        trajectory["type_names"].append([])
        trajectory["positions"].append([])
        edge_ids = []
        edge_positions = []
        for particle_id in monomers.get("particles", {}):
            particle = monomers["particles"][particle_id]
            trajectory["unique_ids"][time_index].append(int(particle_id))
            trajectory["type_names"][time_index].append(particle["type_name"])
            trajectory["positions"][time_index].append(
                np.array(particle["position"]) + position_offset
            )
            # visualize edges between particles
            for neighbor_id in particle["neighbor_ids"]:
                edge = (particle_id, neighbor_id)
                reverse_edge = (neighbor_id, particle_id)
                if edge not in edge_ids and reverse_edge not in edge_ids:
                    edge_ids.append(edge)
                    edge_positions.append(
                        (np.array(particle["position"]) + position_offset).tolist()
                        + (np.array(monomers["particles"][neighbor_id]["position"])
                        + position_offset).tolist()
                    )
        n_agents = len(trajectory["unique_ids"][time_index])
        n_edges = len(edge_ids)
        trajectory["n_agents"].append(n_agents + n_edges)
        trajectory["viz_types"].append(n_agents * [1000.0])
        trajectory["viz_types"][time_index] += n_edges * [1001.0]
        trajectory["unique_ids"][time_index] += [1000 + i for i in range(n_edges)]
        trajectory["type_names"][time_index] += ["edge" for edge in range(n_edges)]
        trajectory["positions"][time_index] += n_edges * [[0.0, 0.0, 0.0]]
        trajectory["radii"].append(n_agents * [actin_radius])
        trajectory["radii"][time_index] += n_edges * [1.0]
        trajectory["n_subpoints"].append(n_agents * [0])
        trajectory["n_subpoints"][time_index] += n_edges * [6]
        trajectory["subpoints"].append(n_agents * [6 * [0.0]])
        trajectory["subpoints"][time_index] += edge_positions
        return trajectory

    @staticmethod
    def fill_df(df, fill):
        """
        Fill Nones in a DataFrame with a fill value
        """
        # Create a dataframe of fill values
        fill_array = [[fill] * df.shape[1]] * df.shape[0]
        fill_df = pd.DataFrame(fill_array)
        # Replace all entries with None with the fill
        df[df.isna()] = fill_df
        return df

    @staticmethod
    def jagged_3d_list_to_numpy_array(jagged_3d_list, length_per_item=3):
        """
        Shape a jagged list with 3 dimensions to a numpy array
        """
        df = SimulariumEmitter.fill_df(pd.DataFrame(jagged_3d_list), length_per_item * [0.0])
        df_t = df.transpose()
        exploded = [df_t[col].explode() for col in list(df_t.columns)]
        result = np.array(exploded).reshape((df.shape[0], df.shape[1], length_per_item))
        return result

    @staticmethod
    def get_agent_data_from_jagged_lists(trajectory, scale_factor) -> AgentData:
        """
        Shape a dictionary of jagged lists into a Simularium AgentData object
        """
        return AgentData(
            times=np.arange(len(trajectory["times"])),
            n_agents=np.array(trajectory["n_agents"]),
            viz_types=SimulariumEmitter.fill_df(
                pd.DataFrame(trajectory["viz_types"]), 1000.0
            ).to_numpy(),
            unique_ids=SimulariumEmitter.fill_df(
                pd.DataFrame(trajectory["unique_ids"]), 0
            ).to_numpy(dtype=int),
            types=trajectory["type_names"],
            positions=scale_factor
            * SimulariumEmitter.jagged_3d_list_to_numpy_array(trajectory["positions"]),
            radii=scale_factor
            * SimulariumEmitter.fill_df(
                pd.DataFrame(trajectory["radii"]), 0.0
            ).to_numpy(),
            n_subpoints=SimulariumEmitter.fill_df(
                pd.DataFrame(trajectory["n_subpoints"]), 0
            ).to_numpy(dtype=int),
            subpoints=scale_factor
            * SimulariumEmitter.jagged_3d_list_to_numpy_array(trajectory["subpoints"], 6),
        )

    @staticmethod
    def get_simularium_converter(
        trajectory, box_dimensions, scale_factor
    ) -> TrajectoryConverter:
        """
        Shape a dictionary of jagged lists into a Simularium TrajectoryData object
        and provide it to a TrajectoryConverter for conversion
        """
        spatial_units = UnitData("nm")
        spatial_units.multiply(1 / scale_factor)
        return TrajectoryConverter(
            TrajectoryData(
                meta_data=MetaData(
                    box_size=scale_factor * box_dimensions,
                ),
                agent_data=SimulariumEmitter.get_agent_data_from_jagged_lists(
                    trajectory, scale_factor
                ),
                time_units=UnitData("count"),
                spatial_units=spatial_units,
            )
        )

    def save_simularium_file(self, output_path: str="test") -> str:
        """
        Save the accumulated timeseries history of emitted data to file
        """
        actin_radius = 3.0
        box_dimensions = np.array(3 * [150.])
        trajectory = {
            "times": [],
            "n_agents": [],
            "viz_types": [],
            "unique_ids": [],
            "type_names": [],
            "positions": [],
            "radii": [],
            "n_subpoints": [],
            "subpoints": [],
        }
        for time, state in self.saved_data.items():
            trajectory = self.get_simularium_monomers(
                time,
                state,
                actin_radius,
                np.zeros(3),
                trajectory,
            )
        simularium_converter = SimulariumEmitter.get_simularium_converter(
            trajectory, box_dimensions, 0.1
        )
        simularium_converter.save(output_path)
        return f"saved to {output_path}.simularium"


import random
from typing import Any

import numpy as np
from process_bigraph import ProcessTypes, Composite
from process_bigraph.emitter import emitter_from_wires, gather_emitter_results
from simularium_readdy_models.actin import ActinGenerator, FiberData
from simularium_readdy_models.common import get_membrane_monomers

from pb_multiscale_actin.processes import ReaddyActinMembrane
from pb_multiscale_actin.processes import SimulariumEmitter


def get_config() -> dict[str, Any]:
    return {
        "name": "actin_membrane",
        "internal_timestep": 0.1,  # ns
        "box_size": np.array([float(150.0)] * 3),  # nm
        "periodic_boundary": True,
        "reaction_distance": 1.0,  # nm
        "n_cpu": 4,
        "only_linear_actin_constraints": True,
        "reactions": True,
        "dimerize_rate": 1e-30,  # 1/ns
        "dimerize_reverse_rate": 1.4e-9,  # 1/ns
        "trimerize_rate": 2.1e-2,  # 1/ns
        "trimerize_reverse_rate": 1.4e-9,  # 1/ns
        "pointed_growth_ATP_rate": 2.4e-5,  # 1/ns
        "pointed_growth_ADP_rate": 2.95e-6,  # 1/ns
        "pointed_shrink_ATP_rate": 8.0e-10,  # 1/ns
        "pointed_shrink_ADP_rate": 3.0e-10,  # 1/ns
        "barbed_growth_ATP_rate": 1e30,  # 1/ns
        "barbed_growth_ADP_rate": 7.0e-5,  # 1/ns
        "nucleate_ATP_rate": 2.1e-2,  # 1/ns
        "nucleate_ADP_rate": 7.0e-5,  # 1/ns
        "barbed_shrink_ATP_rate": 1.4e-9,  # 1/ns
        "barbed_shrink_ADP_rate": 8.0e-9,  # 1/ns
        "arp_bind_ATP_rate": 2.1e-2,  # 1/ns
        "arp_bind_ADP_rate": 7.0e-5,  # 1/ns
        "arp_unbind_ATP_rate": 1.4e-9,  # 1/ns
        "arp_unbind_ADP_rate": 8.0e-9,  # 1/ns
        "barbed_growth_branch_ATP_rate": 2.1e-2,  # 1/ns
        "barbed_growth_branch_ADP_rate": 7.0e-5,  # 1/ns
        "debranching_ATP_rate": 1.4e-9,  # 1/ns
        "debranching_ADP_rate": 7.0e-5,  # 1/ns
        "cap_bind_rate": 2.1e-2,  # 1/ns
        "cap_unbind_rate": 1.4e-9,  # 1/ns
        "hydrolysis_actin_rate": 1e-30,  # 1/ns
        "hydrolysis_arp_rate": 3.5e-5,  # 1/ns
        "nucleotide_exchange_actin_rate": 1e-5,  # 1/ns
        "nucleotide_exchange_arp_rate": 1e-5,  # 1/ns
        "verbose": False,
        "use_box_actin": True,
        "use_box_arp": False,
        "use_box_cap": False,
        "obstacle_radius": 0.0,
        "obstacle_diff_coeff": 0.0,
        "use_box_obstacle": False,
        "position_obstacle_stride": 0,
        "displace_pointed_end_tangent": False,
        "displace_pointed_end_radial": False,
        "tangent_displacement_nm": 0.0,
        "radial_displacement_radius_nm": 0.0,
        "radial_displacement_angle_deg": 0.0,
        "longitudinal_bonds": True,
        "displace_stride": 1,
        "bonds_force_multiplier": 0.2,
        "angles_force_constant": 1000.0,
        "dihedrals_force_constant": 1000.0,
        "actin_constraints": True,
        "use_box_actin": True,
        "actin_box_center_x": 12.0,
        "actin_box_center_y": 0.0,
        "actin_box_center_z": 0.0,
        "actin_box_size_x": 20.0,
        "actin_box_size_y": 50.0,
        "actin_box_size_z": 50.0,
        "add_extra_box": False,
        "barbed_binding_site": True,
        "binding_site_reaction_distance": 3.0,
        "add_membrane": True,
        "membrane_center_x": 25.0,
        "membrane_center_y": 0.0,
        "membrane_center_z": 0.0,
        "membrane_size_x": 0.0,
        "membrane_size_y": 100.0,
        "membrane_size_z": 100.0,
        'membrane_particle_radius': 2.5,
        'obstacle_controlled_position_x': 0.0,
        'obstacle_controlled_position_y': 0.0,
        'obstacle_controlled_position_z': 0.0,
        'random_seed': 0,
    }


def get_monomers():
    actin_monomers = ActinGenerator.get_monomers(
        fibers_data=[
            FiberData(
                28,
                [
                    np.array([-25, 0, 0]),
                    np.array([25, 0, 0]),
                ],
                "Actin-Polymer",
            )
        ],
        use_uuids=False,
        start_normal=np.array([0., 1., 0.]),
        longitudinal_bonds=True,
        barbed_binding_site=True,
    )
    actin_monomers = ActinGenerator.setup_fixed_monomers(
        actin_monomers,
        orthogonal_seed=True,
        n_fixed_monomers_pointed=3,
        n_fixed_monomers_barbed=0,
    )
    membrane_monomers = get_membrane_monomers(
        center=np.array([25.0, 0.0, 0.0]),
        size=np.array([0.0, 100.0, 100.0]),
        particle_radius=2.5,
        start_particle_id=len(actin_monomers["particles"].keys()),
        top_id=1
    )
    free_actin_monomers = ActinGenerator.get_free_actin_monomers(
        concentration=500.0,
        box_center=np.array([12., 0., 0.]),
        box_size=np.array([20., 50., 50.]),
        start_particle_id=len(actin_monomers["particles"].keys()) + len(membrane_monomers["particles"].keys()),
        start_top_id=2
    )
    monomers = {
        'particles': {**actin_monomers['particles'], **membrane_monomers['particles']},
        'topologies': {**actin_monomers['topologies'], **membrane_monomers['topologies']}
    }
    monomers = {
        'particles': {**monomers['particles'], **free_actin_monomers['particles']},
        'topologies': {**monomers['topologies'], **free_actin_monomers['topologies']}
    }
    return monomers


def register_items_into_core(core: ProcessTypes):
    particle = {
        'type_name': 'string',
        'position': 'tuple[float,float,float]',
        'neighbor_ids': 'list[integer]',
        '_apply': 'set',
    }
    topology = {
        'type_name': 'string',
        'particle_ids': 'list[integer]',
        '_apply': 'set',
    }
    core.register('topology', topology)
    core.register('particle', particle)

    core.register_process('readdy', ReaddyActinMembrane)
    core.register_process('simularium-emitter', SimulariumEmitter)


def generate_readdy_state(output_dir: str):
    config = get_config()

    # make the simulation
    random.seed(config['random_seed'])
    np.random.seed(config['random_seed'])

    monomers = get_monomers()

    emitters_from_wires = emitter_from_wires({
        'particles': ['particles'],
        'topologies': ['topologies'],
        'global_time': ['global_time']
    }, address='local:simularium-emitter')
    emitters_from_wires["config"]["output_dir"] = output_dir
    state = {
        "emitter": emitters_from_wires,
        'readdy': {
            '_type': 'process',
            'address': 'local:readdy',
            'config': config,
            'inputs': {
                'particles': ['particles'],
                'topologies': ['topologies']
            },
            'outputs': {
                'particles': ['particles'],
                'topologies': ['topologies']
            }
        },
        **monomers
    }
    return state

def run_readdy_actin_membrane(total_time=3):
    state = generate_readdy_state(output_dir="readdy_result")

    core = ProcessTypes()
    register_items_into_core(core)

    sim = Composite({
        "state": state,
    }, core=core)

    # simulate
    sim.run(total_time)  # time in ns

    results = gather_emitter_results(sim)


if __name__ == "__main__":
    run_readdy_actin_membrane()





"""
==================
Template Composite
==================

This is a toy composite that loads in two template processes and combines them.
"""

# TODO: Delete this file before publishing your project.

from vivarium.core.engine import Engine, pp
from vivarium.core.composer import Composer
from vivarium.library.pretty import format_dict
from vivarium.processes.injector import Injector

from template.processes.template_process import Template


class TemplateComposer(Composer):

    defaults = {
        'template1': {
            'parameter1': 2.0,
        },
        'template2': {
            'parameter1': 5.0,
        },
    }

    def __init__(self, config=None):
        super().__init__(config)

    def generate_processes(self, config):
        template1 = Template(self.config['template1'])
        template2 = Template(self.config['template2'])

        return {
            'template1': template1,
            'template2': template2,
        }

    def generate_topology(self, config):
        return {
            'template1': {
                'internal': ('internal_1',),
                'external': ('external_shared', ),
            },
            'template2': {
                'internal': ('internal_2',),
                'external': ('external_shared', ),
            },
        }


def test_template_composite():
    '''Run a simulation of the composite.

    Returns:
        The simulation output.
    '''

    # 1.) Declare the initial state, mirroring the ports structure. May be passed as a parsable argument.
    initial_state = {
        'internal_1': {
            'A': 0.0
        },
        'internal_2': {
            'A': 10.0
        },
        'external_shared': {
            'A': 1.0
        },
    }

    # 2.) Create the simulation run parameters for the simulator
    config = {
        'template1': {
            'parameter1': 4.0,
        },
        'template2': {
            'parameter1': 4.0,
        }
    }

    # 3.) Initialize the composer by passing in a config dict
    template_composer = TemplateComposer(config)

    # 4.) Generate a composite by calling the Composers' generate method
    template_composite = template_composer.generate()

    # 5.) Feed the compsite instance you just created to the vivarium Engine
    sim = Engine(
        composite=template_composite,
        initial_state=initial_state
    )

    # 6.) Call update with that sim object, which calls the next_update method in the implementation you created above using total_time
    total_time = 3
    sim.update(
        interval=total_time
    )

    # 7.) Get the data which is emitted from the sim object.
    data = sim.emitter.get_timeseries()

    # 7a.) Observe the data which is return from running the process:
    print(f'RESULTS: {pp(data)}')
    return data


# run module with python template/processes/template_composite.py
if __name__ == '__main__':
    test_template_composite()



from process_bigraph.process_types import ProcessTypes

from pb_multiscale_actin.processes.readdy_actin_membrane import ReaddyActinMembrane
from pb_multiscale_actin.processes.simularium_emitter import SimulariumEmitter

particle = {
    'type_name': 'string',
    'position': 'tuple[float,float,float]',
    'neighbor_ids': 'list[int]',
    '_apply': 'set',
}

topology = {
    'type_name': 'string',
    'particle_ids': 'list[int]',
    '_apply': 'set',
}

core = ProcessTypes()

# register types
core.register('topology', topology)
core.register('particle', particle)

core.register_process('readdy', ReaddyActinMembrane)
core.register_process('simularium-emitter', SimulariumEmitter)
